# Generic
DEFAULT_EXPORT_FILENAME = "export"

# Metadata
METADATA_FILENAME = "meta"
METADATA_EXT = "json"
METADATA_CLASS_LABEL = "label"

# TimeSeries
TIME_SERIES_VALUES = "values"
TIME_SERIES_CI_LOWER = "ci_lower"
TIME_SERIES_CI_UPPER = "ci_upper"
TIME_SERIES_FILENAME = "data"
TIME_SERIES_EXT = "csv"
TIME_SERIES_LABEL_PREFIX = "label_"

# Model
MODEL_TYPE_UNIVARIATE = "univariate"
MODEL_TYPE_MULTIVARIATE = "multivariate"

# Misc. IO
PICKLE_EXT = "pkl"
